"""
This file provides information of how to build and configure HepMC framework:
https://github.com/AIDASoft/DD4hep
"""

import os
import platform

from edpm.engine import env_gen
from edpm.engine.commands import run, workdir
from edpm.engine.env_gen import Set, Append, Prepend
from edpm.engine.git_cmake_recipe import GitCmakeRecipe



class DD4HEPRecipe(GitCmakeRecipe):
    """Provides data for building and installing HepMC framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        """
        Installs Genfit track fitting framework
        """

        # Set initial values for parent class and self
        super(DD4HEPRecipe, self).__init__('dd4hep')

        self.config['branch'] = 'v01-23'
        self.config['repo_address'] = 'https://github.com/AIDASoft/DD4hep.git'
        self.config['cmake_flags'] = ' -DDD4HEP_USE_GEANT4=ON -DDD4HEP_USE_EDM4HEP=ON -DDD4HEP_USE_LCIO=OFF '

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""

        """Generates environments to be set

        This function is a bit more tricky than usual - RawText are going to be used.
        RawText actually allows to:
         1. wright a custom text for sh and csh environment scripts
         2. and run a python command to set the environment

        For bash and csh one wants to run 'source .../thisdd4hep_only.sh[csh]'
        But also, one must be set the environment inside python.
        When building Geant from scratch and there is no thisdd4hep_only.sh[csh] yet
        update_python_environment() - do this

        """

        install_path = data['install_path']
        bin_path = os.path.join(install_path, 'bin')
        lib_path = os.path.join(install_path, 'lib')        # on some platforms
        include_path = os.path.join(install_path, 'include')        # on some platforms
        cmake_path = os.path.join(install_path, 'cmake')

        # The next is about conda
        # in conda thisdd4hep_only.sh triggers error explaining, that everything is already done in activate
        # so we don't need to put thisdd4hep_only if we acting under conda
        # this is hacky hack
        is_under_conda = 'DD4HEP_INSTALLED_BY_CONDA' in os.environ

        def update_python_environment():
            """Function that will update Geant environment in python build step
            We need this function because we DON'T want to source geant4.sh in python
            """
            env_updates = [
                env_gen.Append('LD_LIBRARY_PATH', lib_path),
                env_gen.Append('CMAKE_PREFIX_PATH', cmake_path),
                env_gen.Append('ROOT_INCLUDE_PATH', include_path),   # So that rootcling could find it
            ]

            if platform.system() == 'Darwin':
                env_updates.append(env_gen.Append('DYLD_LIBRARY_PATH', lib_path))

            for updater in env_updates:
                updater.update_python_env()

        # We just call geant4.sh in different shells
        yield Prepend('PATH', bin_path)  # to make available clhep-config and others

        # We just call thisroot.xx in different shells
        bash_thisdd4hep_path = os.path.join(bin_path, 'thisdd4hep_only.sh')
        bash_text = '\n' \
                    'if [[ -z "$DD4HEP_INSTALLED_BY_CONDA" ]]; then \n' \
                    '   if test -f "{0}"; then \n' \
                    '      source {0}; \n' \
                    '   fi\n' \
                    'fi\n'.format(bash_thisdd4hep_path)


        # in Geant CSH script Geant asks to get a path for geant bin directory
        csh_text = 'echo "WARNING(!) DD4HEP requires bash to setup environment variables'

        yield env_gen.RawText(
            bash_text,
            csh_text,
            update_python_environment
        )

        install_path = data['install_path']
        bin_path = os.path.join(install_path, 'bin')
        yield Prepend('PATH', bin_path)  # to make available clhep-config and others
        yield Set('DD4HEP_DIR', install_path)


        yield Append('LD_LIBRARY_PATH', os.path.join(install_path, 'lib'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu18': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu18': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {}



"""
This file provides information of how to build and configure CERN.ROOT:
https://github.com/root-project/root

"""

import os
from distutils.dir_util import mkpath

from edpm.engine import env_gen
from edpm.engine.db import BUILT_WITH_CONFIG
from edpm.engine.recipe import Recipe
from edpm.engine.commands import run, env, workdir, is_not_empty_dir

ROOTSYS = "ROOTSYS"


class RootRecipe(Recipe):
    """Provides data for building and installing root

    PackageInstallationContext is located in engine/recipe.py

    """

    class DefaultConfigFields(object):
        pass

    def __init__(self):
        """
        """

        # Fill the common path pattern
        super(RootRecipe, self).__init__("root")
        self.config['branch'] = 'v6-26-10'
        self.config['cmake_custom_flags'] = ''
        self.config['cmake_build_type'] = 'RelWithDebInfo'
        self.config['cxx_standard'] = '17'

    def find_python(self):
        """Searches default python which is first found in PATH"""
        from subprocess import check_output
        out = check_output(["which", "python3"]).decode('ascii').strip()

        if not out:
            out = check_output(["which", "python2"]).decode('ascii').strip()

        if not out:
            out = check_output(["which", "python"]).decode('ascii').strip()
        return out

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        # Ensure that we are using C++14 or higher
        if int(self.config['cxx_standard']) < 14:
            message = "ERROR. cxx_standard must be 14 or above to build root7.\n" \
                      "To set cxx_standard globally:\n" \
                      "   edpm config global cxx_standard=14\n" \
                      "To set cxx_standard for acts:\n" \
                      "   edpm config root7 cxx_standard=14\n" \
                      "(!) Make sure cmake is regenerated after. (rm <top_dir>/acts and run edpm install acts again)\n"
            raise ValueError(message)
        #
        # Compile with python3, then whatever python is...
        python_path = self.find_python()
        self.config["python_flag"] = ' -DPYTHON_EXECUTABLE={} '.format(python_path) if python_path else ''
        # >oO debug: print("Configuring ROOT with '{}' python flag".format(self.config["python_flag"]))

        #
        # use_common_dirs_scheme sets standard package variables:
        # version      = 'v{}-{:02}-{:02}'                 # Stringified version. Used to create directories and so on
        # source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
        # build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
        # install_path = {app_path}/root-{version}         # Where the binary installation is
        self.use_common_dirs_scheme()

        #
        # Root download link. We will use github root mirror:
        # The tags have names like: v6-14-04
        # http://github.com/root-project/root.git
        # clone with shallow copy
        self.clone_command = "git clone --depth 1 -b {branch} https://github.com/root-project/root.git {source_path}" \
            .format(**self.config)

        # Make sure custom flags are in there
        if "cmake_custom_flags" not in self.config:
            self.config["cmake_custom_flags"] = ''

        #
        # ROOT packets to disable in our build (go with -D{name}=ON flag)
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        self.build_cmd = "cmake -Wno-dev -DCMAKE_INSTALL_PREFIX={install_path} " \
                         " -DCMAKE_CXX_STANDARD={cxx_standard}" \
                         " -DCMAKE_BUILD_TYPE={cmake_build_type}"\
                         " -Dhttp=ON" \
                         " -Droot7=ON" \
                         " -Dgdml=ON" \
                         " -Dxrootd=OFF" \
                         " -Dmysql=OFF" \
                         " -Dpythia6=OFF" \
                         " -Dpythia6_nolink=OFF" \
                         " -Dpythia8=OFF" \
                         " -Dminuit2=ON" \
                         " {python_flag}" \
                         " {cmake_custom_flags}" \
                         " {source_path}" \
                         "&& cmake --build . -- -j {build_threads}" \
                         "&& cmake --build . --target install" \
            .format(**self.config)  # make global options like '-j8'. Skip now

    def step_install(self):
        self.step_clone_root()
        self.step_build_root()

    def step_clone_root(self):
        """Clones root from github mirror"""

        # Check the directory exists and not empty
        if is_not_empty_dir(self.source_path):
            return  # The directory exists and is not empty. Assume it cloned

        mkpath(self.source_path)  # Create the directory and any missing ancestor directories if not there
        run(self.clone_command)  # Execute git clone command

    def step_build_root(self):
        """Builds root from the ground"""

        # Create build directory
        run('mkdir -p {}'.format(self.build_path))

        env('ROOTSYS', self.install_path)

        # go to our build directory
        workdir(self.build_path)

        # run cmake && make && install
        run(self.build_cmd)

    def step_rebuild_root(self):
        """Clear root build directory"""

        # clear sources directories if needed
        run('rm -rf {}'.format(self.source_path))
        run('rm -rf {}'.format(self.build_path))

        # Now run build root
        self.step_build_root()

    @staticmethod
    def gen_env(data):
        install_path = data['install_path']
        yield env_gen.Prepend('CMAKE_PREFIX_PATH', os.path.join(install_path, 'cmake/'))

        isinstance(data, dict)

        # The next is about conda
        # in conda thisroot.sh triggers error explaining, that everything is already done in activate
        # so we don't need to put thisroot if we acting under conda
        # this is hacky hack
        is_under_conda = 'ROOT_INSTALLED_BY_CONDA' in os.environ

        # In any case we need python environment to build stuff with root under edpm
        def update_python_environment():
            """Function that will update ROOT environment in python
            We need this function because we DON'T want source thisroot in python
            """

            root_bin = os.path.join(install_path, 'bin')
            root_lib = os.path.join(install_path, 'lib')
            root_jup = os.path.join(install_path, 'etc', 'notebook')

            env_updates = [
                env_gen.Set('ROOTSYS', install_path),
                env_gen.Prepend('PATH', root_bin),
                env_gen.Prepend('LD_LIBRARY_PATH', root_lib),
                env_gen.Prepend('DYLD_LIBRARY_PATH', root_lib),
                env_gen.Prepend('PYTHONPATH', root_lib),
                env_gen.Prepend('CMAKE_PREFIX_PATH', install_path),
                env_gen.Prepend('JUPYTER_PATH', root_jup),
            ]

            for updater in env_updates:
                updater.update_python_env()

        # We just call thisroot.xx in different shells
        bash_thisroot_path = os.path.join(install_path, 'bin', 'thisroot.sh')
        bash_text = '\n' \
                    'if [[ -z "$ROOT_INSTALLED_BY_CONDA" ]]; then \n' \
                    '   if test -f "{0}"; then \n' \
                    '      source {0}; \n' \
                    '   fi\n' \
                    'fi\n'.format(bash_thisroot_path)

        csh_thisroot_path = os.path.join(install_path, 'bin', 'thisroot.csh')
        csh_text = "\n" \
                   "if (! $?ROOT_INSTALLED_BY_CONDA) then\n" \
                   "   if ( -f {0} ) then\n"\
                   "      source {0}\n"\
                   "   endif\n" \
                   "endif\n"\
                   .format(csh_thisroot_path)

        bash_text = bash_text if not is_under_conda else "# Don't call thisroot.sh under conda"
        csh_text = csh_text if not is_under_conda else "# Don't call thisroot.csh under conda"
        # Do we need this because of conda?

        # RawText requires text for bash, csh and a function for python
        raw = env_gen.RawText(bash_text, csh_text, update_python_environment)
        yield raw

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    #                   }
    os_dependencies = {
        'required': {
            'ubuntu18': "dpkg-dev binutils libx11-dev libxpm-dev libxft-dev libxext-dev liblzma-dev",
            'ubuntu22': "dpkg-dev binutils libx11-dev libxpm-dev libxft-dev libxext-dev liblzma-dev",
            'centos7': "gcc binutils libX11-devel libXpm-devel libXft-devel libXext-devel",
            'centos8': "gcc binutils libX11-devel libXpm-devel libXft-devel libXext-devel"
        },
        'optional': {
            'ubuntu18': "gfortran libssl-dev libpcre3-dev "
                      "xlibmesa-glu-dev libglew-dev libftgl-dev "
                      "libmysqlclient-dev libfftw3-dev libcfitsio-dev "
                      "graphviz-dev libavahi-compat-libdnssd-dev "
                      "libldap2-dev python3-dev libxml2-dev libkrb5-dev "
                      "libgsl0-dev",
            'ubuntu22': "gfortran libssl-dev libpcre3-dev "
                        "xlibmesa-glu-dev libglew-dev libftgl-dev "
                        "libmysqlclient-dev libfftw3-dev libcfitsio-dev "
                        "graphviz-dev libavahi-compat-libdnssd-dev "
                        "libldap2-dev python3-dev libxml2-dev libkrb5-dev "
                        "libgsl0-dev",

            'centos7': "gcc-gfortran openssl-devel pcre-devel "
                      "mesa-libGL-devel mesa-libGLU-devel glew-devel ftgl-devel mysql-devel "
                      "fftw-devel cfitsio-devel graphviz-devel "
                      "avahi-compat-libdns_sd-devel libldap-dev python-devel "
                      "libxml2-devel gsl-static",
            'centos8': "gcc-gfortran openssl-devel pcre-devel "
                      "mesa-libGL-devel mesa-libGLU-devel ftgl-devel mysql-devel "
                      "fftw-devel cfitsio-devel graphviz "
                      "openldap-devel python3-devel "
                      "libxml2-devel gsl-devel"
        },
    }




def root_find():
    """Looks for CERN ROOT package
    :return [str] - empty list if not found or a list with 1 element - ROOTSYS path

    The only way to find ROOT is by checking for ROOTSYS package,
    The function family xxx_find() return list in general
    so this function returns either empty list or a list with 1 element - root path
    """

    # The only way to find a CERN ROOT is by
    result = []

    # Check ROOTSYS environment variable
    if ROOTSYS not in os.environ:
        print("<red>ROOTSYS</red> not found in the environment")
        return result

    # Now check ROOTSYS exists in the system
    root_sys_path = os.environ[ROOTSYS]
    if not os.path.isdir(root_sys_path):
        print("WARNING", " ROOTSYS points to nonexistent directory of a file")
        return result

    # Looks like root exists, return the path
    return [root_sys_path]



"""
EPIC DD4Hep geometry repository
"""

import os

from edpm.engine.env_gen import Set, Append, Prepend
from edpm.engine.git_cmake_recipe import GitCmakeRecipe


class EpicRecipe(GitCmakeRecipe):


    def __init__(self):

        # Set initial values for parent class and self
        super(EpicRecipe, self).__init__('epic')    # This name will be used in edpm commands
        self.required_deps = ['clhep', 'eigen3', 'root', 'hepmc3', 'podio', 'edm4hep', 'edm4eic', 'geant4','dd4hep', 'acts']
        self.config['branch'] = 'main'                             # The branch or tag to be cloned (-b flag)
        self.config['repo_address'] = 'https://github.com/eic/epic'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        path = data['install_path']

        # it could be lib or lib64. There are bugs on different platforms (RHEL&centos and WSL included)
        # https://stackoverflow.com/questions/46847939/config-site-for-vendor-libs-on-centos-x86-64
        # https: // bugzilla.redhat.com / show_bug.cgi?id = 1510073

        # yield Prepend('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake'))
        yield Prepend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield Prepend('PATH', os.path.join(path, 'bin'))
        yield Set('DETECTOR_PATH', os.path.join(path, 'share', 'epic'))
        yield Set('BEAMLINE', 'epic')
        yield Set('BEAMLINE_PATH', os.path.join(path, 'share', 'epic'))
        yield Set('BEAMLINE_CONFIG', 'epic')

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {},
        'optional': {},
    }


import os
import click

from edpm.engine.api import pass_edpm_context, DB_FILE_PATH, ENV_CSH_PATH, ENV_SH_PATH, edpmApi, print_packets_info
from edpm.engine.db import PacketStateDatabase
from edpm.engine.output import markup_print as mprint


def print_first_time_message():
    mprint("""
The database file doesn't exist. Probably you run 'edpm' for one of the first times.

1. Install or check OS maintained required packages:
    > edpm req ubuntu         # for all packets edpm knows to built/install
    > edpm req ubuntu eicrecon   # for eicrecon and its dependencies only
   
   * - at this point put 'ubuntu' for debian and 'centos' for RHEL and CentOS systems. 
   Will be updated in future to support macOS, and to have grained versions

1. Set <b><blue>top-dir</blue></b> to start. This is where all missing packets will be installed.   

   > edpm --top-dir=<where-to-install-all>
   
2. You may have CERN.ROOT installed (req. version >= 6.14.00). Run this:

   > edpm set root `$ROOTSYS`
   
   You may set paths for other installed dependencies:
   > edpm install eicrecon --missing --explain    # to see missing dependencies
   > edpm set <name> <path>                    # to set dependency path
   
3. Then you can install all missing dependencies:

   > edpm install eicrecon --missing
   

P.S - you can read this message by adding --help-first flag
    - edpm gitlab: https://gitlab.com/eic/edpm
    - This message will disappear after running any command that make changes
    """)
    click.echo()


_starting_workdir = ""

@click.group(invoke_without_command=True)
@click.option('--debug/--no-debug', default=False)
@click.option('--top-dir', default="")
@pass_edpm_context
@click.pass_context
def edpm_cli(ctx, ectx, debug, top_dir):
    """edpm stands for EIC Development Packet Manager"""

    assert isinstance(ectx, edpmApi)    # Type check for ectx

    # Load db and modules from disk
    db_existed = ectx.load_shmoad_ugly_toad()    # False => Couldn't load and used default

    # user asks to set the top dir
    if top_dir:
        ectx.db.top_dir = os.path.abspath(os.path.normpath(top_dir))
        ectx.db.save()
        db_existed = True

    # check if DB file already exists
    if not db_existed:
        print_first_time_message()
    else:
        # if there is no commands and we loaded the DB lets print some info:
        if ctx.invoked_subcommand is None:
            from edpm.version import version
            mprint("<b><blue>edpm</blue></b> v{}".format(version))
            mprint("<b><blue>top dir :</blue></b>\n  {}", ectx.db.top_dir)
            mprint("<b><blue>state db :</blue></b>\n  {}", ectx.config[DB_FILE_PATH])
            mprint("  (users are encouraged to inspect/edit it)")
            mprint("<b><blue>env files :</blue></b>\n  {}\n  {}", ectx.config[ENV_SH_PATH], ectx.config[ENV_CSH_PATH])
            print_packets_info(ectx.db)

from edpm.cli.env import env as env_group
from edpm.cli.install import install as install_group
from edpm.cli.find import find as find_group
from edpm.cli.req import req as requirements_command
from edpm.cli.set import set as set_command
from edpm.cli.rm import rm as rm_command
from edpm.cli.pwd import pwd as pwd_command
from edpm.cli.clean import clean as clean_command
from edpm.cli.info import info as info_command
from edpm .cli.config import config as config_command
from edpm .cli.mergedb import mergedb as mergedb_command

edpm_cli.add_command(install_group)
edpm_cli.add_command(find_group)
edpm_cli.add_command(env_group)
edpm_cli.add_command(requirements_command)
edpm_cli.add_command(set_command)
edpm_cli.add_command(rm_command)
edpm_cli.add_command(pwd_command)
edpm_cli.add_command(clean_command)
edpm_cli.add_command(info_command)
edpm_cli.add_command(config_command)
edpm_cli.add_command(mergedb_command)

if __name__ == '__main__':
    edpm_cli()



import click


# A list of known text names of colors
known_colors = ["black",        # (might be a gray)
                "red",
                "green",
                "yellow",       # (might be an orange)
                "blue",
                "magenta",
                "cyan",
                "white",        # (might be light gray)
                "bright_black",
                "bright_red",
                "bright_green",
                "bright_yellow",
                "bright_blue",
                "bright_magenta",
                "bright_cyan",
                "bright_white",
                "reset"]


def color_echo(title, text, color, title_bold=False, text_bold=False, nl=True):
    """
    Prints colorful title and then text in one line

    :param title:      text in color
    :param text:       later text without color
    :param color:      color of the title
    :param title_bold: title is bold
    :param text_bold:  text is bold
    :param nl:         new line after the text
    """
    click.secho(title, fg=color, bold=title_bold, nl=False)
    click.secho(text, bold=text_bold, nl=nl)


# noinspection PyBroadException
try:
    # noinspection PyUnresolvedReferences
    from HTMLParser import HTMLParser
except:
    from html.parser import HTMLParser


class ColorMarkupClickPrinter(HTMLParser, object):

    def error(self, message):
        pass

    is_bold = False       # the next printed text will be bold
    is_underline = False  # the next printed text will be underlined
    color = ""            # color of the next printed text

    def handle_starttag(self, tag, attrs):

        # make in case independent
        tag = tag.lower()

        # do we know this tag?
        if tag == 'b':
            self.is_bold = True
        elif tag == 'u':
            self.is_underline = True
        elif tag in known_colors:
            self.color = tag
        else:
            click.secho('<{}>'.format(tag), fg=self.color, bold=self.is_bold, underline=self.is_underline, nl=False)


    def handle_endtag(self, tag):
        # make in case independent
        tag = tag.lower()

        # do we know this tag?
        if tag == 'b':
            self.is_bold = False
        elif tag == 'u':
            self.is_underline = False
        elif tag in known_colors:
            self.color = "reset"
        else:
            click.secho('</{}>'.format(tag), fg=self.color, bold=self.is_bold, underline=self.is_underline, nl=False)

    def handle_data(self, data):
        click.secho(data, fg=self.color, bold=self.is_bold, underline=self.is_underline, nl=False)


def markup_print(text, *args, **kwargs):
    """markdown print
    <red> </red>
    <green> <green>
    <blue><blue>
    """
    if args or kwargs:
        text = text.format(*args, **kwargs)
    parser = ColorMarkupClickPrinter()
    parser.feed(text)
    print("")     # new line in the end


if __name__ == "__main__":
    markup_print("One two <b> df <red> my red red red </red> </b>")


import click

from edpm.engine.api import pass_edpm_context, edpmApi
from edpm.engine.output import markup_print as mprint


_help_option_db = "Removes only DB record"
_help_option_all = "Removes DB record and packet folder from disk"
_help_option_auto = "Removes from DB and disk if(!) the packet is owned by edpm"


# @click.group(invoke_without_command=True)
@click.command()
@click.argument('packet_name', nargs=1, metavar='<packet-name>')
@click.argument('install_paths', nargs=-1, metavar='<path>')
@click.option('--db', 'mode', flag_value='db', help=_help_option_db)
@click.option('--all', 'mode', flag_value='all', help=_help_option_all)
@click.option('--auto', 'mode', flag_value='auto', help=_help_option_auto, default=True)
@pass_edpm_context
@click.pass_context
def rm(ctx, ectx, packet_name, install_paths, mode):
    """Removes a packet.
    By default deletes record from edpm DB and the disk folders if the packet is 'owned' by edpm.

    Usage:
        edpm rm <packet-name>         # removes active install of the packet
        edpm rm <packet-name> <path>  # removes the install with the path

    """
    from edpm.engine.db import INSTALL_PATH, IS_OWNED, SOURCE_PATH, BUILD_PATH
    assert isinstance(ectx, edpmApi)

    # We need DB ready for this cli command
    ectx.ensure_db_exists()

    # Check that the packet name is from known packets
    ectx.ensure_installer_known(packet_name)

    if not install_paths:
        install_data = ectx.db.get_active_install(packet_name)
        if not install_data:
            print("No active installation data found for the packet {}".format(packet_name))
            raise click.Abort()
        else:
            print("No path provided. <b>Using 'active' install</b>")
    else:
        install_path = install_paths[0]     # todo multiple paths
        install_data = ectx.db.get_install(packet_name, install_path)
        if not install_data:
            print("No active installation data found for the packet {} and path:\n{}"
                  .format(packet_name, install_path))
            raise click.Abort()

    mprint("<blue><b>Removing install with path: </b></blue>")
    mprint("  {}\n", install_data[INSTALL_PATH])

    remove_folder = False
    if mode == 'all':
        remove_folder = True
    if mode == 'auto':
        remove_folder = install_data[IS_OWNED]
        if not remove_folder:
            mprint("<b>(!)</b> the packet is not 'owned' by edpm. The record is removed from DB but\n"
                   "<b>(!)</b>you have to remove the folder manually:\n{}\n", install_data[INSTALL_PATH])

    ectx.db.remove_install(packet_name, install_data[INSTALL_PATH])
    ectx.db.save()

    # Update environment scripts
    mprint("Updating environment script files...\n")
    ectx.save_default_bash_environ()
    ectx.save_default_csh_environ()

    # remove the folder
    if remove_folder:
        mprint("...trying to remove the folder from disk...\n")
        from edpm.engine.commands import run

        run('rm -rf "{}"'.format(install_data[INSTALL_PATH]))

        if SOURCE_PATH in install_data:
            run('rm -rf "{}"'.format(install_data[SOURCE_PATH]))

        if BUILD_PATH in install_data:
            run('rm -rf "{}"'.format(install_data[BUILD_PATH]))



"""
This file provides information of how to build and configure HepMC framework:
https://gitlab.cern.ch/hepmc/HepMC3
"""

import os

from edpm.engine.commands import run, workdir, env
from edpm.engine.env_gen import Set, Append, Prepend
from edpm.engine.git_cmake_recipe import GitCmakeRecipe


class HepMC3Recipe(GitCmakeRecipe):
    """Provides data for building and installing HepMC framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        """
        Installs Genfit track fitting framework
        """

        # Set initial values for parent class and self
        super(HepMC3Recipe, self).__init__('hepmc3')

        self.config['branch'] = '3.2.5'
        self.config['repo_address'] = 'https://gitlab.cern.ch/hepmc/HepMC3.git'
        self.config['cmake_flags'] = ' -DHEPMC3_ENABLE_ROOTIO=OFF -DHEPMC3_ENABLE_PYTHON=OFF -DHEPMC3_BUILD_EXAMPLES=OFF HEPMC3_CXX_STANDARD=17'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""

        install_path = data['install_path']
        bin_path = os.path.join(install_path, 'bin')
        yield Prepend('PATH', bin_path)  # to make available clhep-config and others
        yield Set('HEPMC3_DIR', install_path)


        yield Append('CMAKE_PREFIX_PATH', os.path.join(install_path, 'share', 'HepMC3', 'cmake'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {},
        'optional': {},
    }


"""
This file provides information of how to build and configure Eigen3 packet:
https://github.com/nathanwbrei/phasm

"""

import os

from edpm.engine.env_gen import Prepend, Set, Append
from edpm.engine.git_cmake_recipe import GitCmakeRecipe


class PhasmRecipe(GitCmakeRecipe):
    """Provides data for building and installing Eicgen3 framework"""

    def __init__(self):
        super(PhasmRecipe, self).__init__('phasm')
        self.config['branch'] = 'master'
        self.config['repo_address'] = 'https://github.com/nathanwbrei/phasm'

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        #
        # use_common_dirs_scheme sets standard package variables:
        # source_path  = {app_path} / src   / {branch}       # Where the sources for the current version are located
        # build_path   = {app_path} / build / {branch}       # Where sources are built. Kind of temporary dir
        # install_path = {app_path} / geant-{branch}         # Where the binary installation is
        self.use_common_dirs_scheme()
        print("Sorry Nathan! I haven't put all the deps yet!")
        exit(1)
        super(PhasmRecipe, self).setup()

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        path = data['install_path']

        yield Prepend('CMAKE_PREFIX_PATH', os.path.join(path, 'share/eigen3/cmake/'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {},
        'optional': {},
    }



"""
Algorithms library:
https://github.com/eic/algorithms

"""

import os

import platform
from edpm.engine.env_gen import Set, Append, Prepend
from edpm.engine.git_cmake_recipe import GitCmakeRecipe


class AlgorithmsRecipe(GitCmakeRecipe):

    def __init__(self):

        # Set initial values for parent class and self
        super(AlgorithmsRecipe, self).__init__('algorithms')                 # This name will be used in edpm commands
        self.config['branch'] = 'algorithms-integration'                     # The branch or tag to be cloned (-b flag)
        self.config['repo_address'] = 'https://github.com/eic/algorithms'    # Repo address
        self.config['cmake_flags'] = ' '
        self.config['cxx_standard'] = 17

    def setup(self, db):
        # Call GitCmakeRecipe `default` setup function
        super(AlgorithmsRecipe, self).setup(db)

        # cmake command:
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        self.build_cmd = "cmake -Wno-dev " \
                         "-DCMAKE_INSTALL_PREFIX={install_path} -DCMAKE_CXX_STANDARD={cxx_standard} " \
                         "{cmake_flags} {cmake_custom_flags} {source_path}/external/algorithms" \
                         "&& cmake --build . -- -j {build_threads}" \
                         "&& cmake --build . --target install" \
                         .format(**self.config)

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        install_path = data['install_path']

        if platform.system() == 'Darwin':
            yield Append('DYLD_LIBRARY_PATH', os.path.join(install_path, 'lib'))

        yield Append('LD_LIBRARY_PATH', os.path.join(install_path, 'lib'))


    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu18': "libmsgsl-dev",
            'ubuntu22': "libmsgsl-dev",
            'centos7': "guidelines-support-library",
            'centos8': "guidelines-support-library",
        },
    }



"""
Acts dd4hep project
https://github.com/acts-project/acts-dd4hep
"""

import os

from edpm.engine.env_gen import Set, Append, Prepend
from edpm.engine.git_cmake_recipe import GitCmakeRecipe


class ActsDD4Hep(GitCmakeRecipe):
    """Provides Recipe for building Acts-DD4Hep plugin connecting library and setting its environment"""

    def __init__(self):
        """
        Installs Acts-DD4Hep plugin connecting library
        """

        # Set initial values for parent class and self
        super(ActsDD4Hep, self).__init__('actsdd4hep')             # This name will be used in edpm commands
        self.config['branch'] = 'v1.0.1'                           # The branch or tag to be cloned (-b flag)
        self.required_deps = ['dd4hep', 'acts']
        self.config['repo_address'] = 'https://github.com/acts-project/acts-dd4hep.git'      # Repo address
        self.config['cxx_standard'] = 17

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        path = data['install_path']

        import platform
        if platform.system() == 'Darwin':
            yield Append('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        yield Append('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield Append('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake', 'ActsDD4hep'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {}




import click

from edpm.engine.api import pass_edpm_context, edpmApi
from edpm.engine.output import markup_print as mprint

# @click.group(invoke_without_command=True)
@click.command()
@click.argument('os_name', nargs=1, metavar='<os-name>')
@click.argument('args', nargs=-1, metavar='<packet-names>')
@click.option('--optional', 'print_mode', flag_value='optional', help="Print optional packages")
@click.option('--required', 'print_mode', flag_value='required', help="Print required packages")
@click.option('--all', 'print_mode', flag_value='all', help="Print all packages (ready for packet manager install)")
@click.option('--all-titles', 'print_mode', flag_value='all_titles', help="Print all packages (human readable)", default=True)
@pass_edpm_context
@click.pass_context
def req(ctx, ectx, os_name, args, print_mode):
    """req - Shows required packages that can be installed by operating system.

    \b
    Example:
      req ubuntu22
      req centos7 eicrecon
      req centos8 root clhep

    By adding --optional, --required, --all flags you can use this command with packet managers:\n
      apt install `edpm req ubuntu --all`

    Known OS: ubuntu18 ubuntu22 centos7 centos8

    Aliases: rhel, rhel7, rhel8, mint, debian, ubuntu
    """

    os_aliases = {
        "ubuntu": "ubuntu22",
        "rhel": "centos8",
        "rhel7": "centos7",
        "rhel8": "centos8",
        "mint": "ubuntu22",
        "debian": "ubuntu18"
    }

    assert isinstance(ectx, edpmApi)

    # We need DB ready for this cli command
    ectx.ensure_db_exists()

    # We have some args, first is os name like 'ubuntu' or 'centos'
    known_os = ectx.req_get_known_os()

    if os_name in os_aliases.keys():
        os_name = os_aliases[os_name]

    if os_name not in known_os:
        mprint('<red><b>ERROR</b></red>: name "{}" is unknown\nKnown os names are:', os_name)
        for name in known_os:
            mprint('   {}', name)
        click.echo(ctx.get_help())
        ctx.exit(1)

    # We have something like 'ubuntu eicrecon'
    required, optional = ectx.req_get_deps(os_name, args)

    if print_mode == "optional":
        mprint(" ".join(optional))
    elif print_mode == "required":
        mprint(" ".join(required))
    elif print_mode == "all":
        mprint(" ".join(optional + required))
    else:
        # print all with juman readable titles
        mprint("<blue><b>REQUIRED</b></blue>:")
        mprint(" ".join(required))
        mprint("<blue><b>OPTIONAL</b></blue>:")
        mprint(" ".join(optional))



# this file contain some snippets and shortcuts to bring compatibility between python 2 and 3
# and not to jump into six or __futures

try:
    # noinspection PyUnresolvedReferences
    to_unicode = unicode
except NameError:
    to_unicode = str



import click
from edpm.engine.api import pass_edpm_context, edpmApi
from edpm.engine.recipe import Recipe
from edpm.engine.output import markup_print as mprint


@click.command()
@click.argument('name_values', nargs=-1)
@pass_edpm_context
def config(ectx, name_values):
    """Sets build config for a packet

    If packet name is put, config goes into that packet. If no packet name is provided,
    config goes into global_config which affects all packet installations.

    Example:
        build_threads=4 jana branch=greenfield

    Explanation: global parameter 'build_threads' is set to 4, and 'jana' parameter  branch is set to 'greenfield'

    The example above is an extreme use case of this command and it is advised to split the contexts:
    >> edpm config build_threads=4
    >> edpm config jana branch=greenfield
    """

    assert isinstance(ectx, edpmApi)

    # We need DB ready for this cli command
    ectx.load_db_if_exists()

    # We need at least some base configuration of recipes
    ectx.configure_recipes()

    if len(name_values) > 1:
        _set_configs(ectx, name_values)
    elif len(name_values) == 1:
        _show_configs(ectx, name_values[0])
    else:
        _show_configs(ectx, 'global')


def _show_configs(ectx, name):
    # get existing config
    if name == 'global':
        build_config = ectx.db.get_global_config()
    else:
        ectx.ensure_installer_known(name)
        build_config = ectx.db.get_config(name)

    mprint('<b><magenta>{}</magenta></b>:'.format(name))                      # pretty printing
    for param_name, value in build_config.items():
        mprint(' <b><blue>{}</blue></b>: {}'.format(param_name, value))

    # There is nothing more than 'global'
    if name == 'global':
        return

    mprint('<b><magenta>Default configs for {}</magenta></b>:'.format(name))  # pretty printing
    recipe = ectx.pm.recipes_by_name[name]
    assert isinstance(recipe, Recipe)

    for param_name, value in recipe.config.items():
        mprint(' <b><blue>{}</blue></b>: {}'.format(param_name, value))


def _set_configs(ectx, name_values):
    # Check that the packet name is from known packets
    config_blob = _process_name_values(name_values)

    for context_name in config_blob.keys():
        mprint('<b><magenta>{}</magenta></b>:'.format(context_name))  # pretty printing

        # get existing config
        if context_name == 'global':
            existing_config = ectx.db.get_global_config()
        else:
            ectx.ensure_installer_known(context_name)
            existing_config = ectx.db.get_config(context_name)

        updating_config = config_blob[context_name]

        existing_config.update(updating_config)  # update config
        ectx.db.save()  # save config

        # pretty printing the updated config
        for name, value in existing_config.items():
            mprint(' <b><blue>{}</blue></b>: {}'.format(name, value))


def _process_name_values(name_values):
    """Converts input parameters to a config map

    >>> _process_name_values(['build_threads=4', 'jana', 'branch=greenfield',  'build_threads=1'])
    {'global': {'build_threads': '4'}, 'jana': {'branch': 'greenfield', 'build_threads': '1'}}

    """

    context = 'global'
    result = {context: {}}

    for name_value in name_values:
        if '=' in name_value:
            name, value = tuple(name_value.split('=', 1))  # 1 as we want to split only the first occurrence
            result[context][name] = value
        else:
            context = name_value
            keys = list(result)
            if context not in keys:
                result[context] = {}

    # remove empty records:
    keys = list(result)   # to avoid RuntimeError: dictionary changed size during iteration
    for key in keys:
        if not result[key]:
            del result[key]

    return result



import os
import inspect
from setuptools import setup
from edpm.version import version as edpm_version

# The directory containing this file
this_script_dir = os.path.dirname(inspect.stack()[0][1])

# The text of the README file
with open(os.path.join(this_script_dir, "pip_readme.md"), 'r') as readme_file:
    readme = readme_file.read()

# This call to setup() does all the work
setup(
    name="edpm",
    version=edpm_version,
    description="EIC Development Package Manager",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/eic/edpm",
    author="Dmitry Romanov",
    author_email="romanov@jlab.org",
    license="MIT",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=["edpm"],
    include_package_data=True,
    setup_requires=["click", "appdirs"],
    install_requires=["click", "appdirs"],
    entry_points={
        "console_scripts": [
            "edpm=edpm:edpm_cli",
        ]
    },
)


# Info - prints extended information of the edpm state

import click

from edpm.engine.api import pass_edpm_context, edpmApi
from edpm.engine.db import INSTALL_PATH
from edpm.engine.output import markup_print as mprint


_cmake_opt_help = "List packages in terms of CMake flags"
_flag_help_db = "Prints information about edpm DB"
_flag_help_db_path = "Prints edpm json DB path"


def _no_flags_set(flag_cmake, flag_db, flag_db_path):
    return flag_cmake and flag_db and flag_db_path


@click.command()
@click.option('--cmake', 'flag_cmake', flag_value='cmake', help=_cmake_opt_help)
@click.option('--db', 'flag_db', flag_value='cmake', help=_cmake_opt_help)
@click.option('--db-path', 'flag_db_path', flag_value='cmake', help=_cmake_opt_help)
@pass_edpm_context
@click.pass_context
def info(ctx, ectx, flag_cmake, flag_db, flag_db_path):
    """info - Description

    \b
    Example:
      info --cmake
      info --db-path
    """

    if _no_flags_set(flag_cmake, flag_db, flag_db_path):
        flag_db = True

    assert isinstance(ectx, edpmApi)

    # We need DB ready for this cli command
    ectx.ensure_db_exists()

    if flag_cmake:
        _print_cmake(ectx)


def _print_cmake(ectx):
    db = ectx.db
    pm = ectx.pm

    flag_names_by_packet_names = pm.recipes_by_name["eicrecon"].cmake_deps_flag_names

    flags = ['-D{}="{}"'.format(flag_names_by_packet_names[name], install_info[INSTALL_PATH])
             for name, install_info in zip(db.packet_names, map(db.get_active_install, db.packet_names))
             if name in flag_names_by_packet_names.keys() and install_info]

    # Fancy print of installed packets
    print(" ".join(flags))


def _print_installed(db):
    installed_names = [name for name in db.packet_names]

    # Fancy print of installed packets
    if installed_names:
        mprint('\n<b><magenta>INSTALLED PACKETS:</magenta></b> (*-active):')
        for packet_name in installed_names:
            mprint(' <b><blue>{}</blue></b>:'.format(packet_name))
            installs = db.get_installs(packet_name)
            for i, installation in enumerate(installs):
                from edpm.engine.db import IS_OWNED, IS_ACTIVE, INSTALL_PATH

                is_owned_str = '<green>(owned)</green>' if installation[IS_OWNED] else ''
                is_active = installation[IS_ACTIVE]
                is_active_str = '*' if is_active else ' '
                path_str = installation[INSTALL_PATH]
                id_str = "[{}]".format(i).rjust(4) if len(installs) > 1 else ""
                mprint("  {}{} {} {}".format(is_active_str, id_str, path_str, is_owned_str))


"""
This file provides information of how to build and configure ACTS framework:
https://gitlab.cern.ch/acts/acts-core
"""

import os

from edpm.engine.env_gen import Set, Append, Prepend
from edpm.engine.git_cmake_recipe import GitCmakeRecipe


class ActsRecipe(GitCmakeRecipe):
    """Provides data for building and installing Genfit framework
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """

    def __init__(self):
        """
        Installs Genfit track fitting framework
        """

        # Set initial values for parent class and self
        super(ActsRecipe, self).__init__('acts')                        # This name will be used in edpm commands
        self.config['branch'] = 'v19.10.0'                              # The branch or tag to be cloned (-b flag)
        self.config['repo_address'] = 'https://github.com/acts-project/acts'    # Repo address
        self.config['cmake_flags'] = '-DACTS_BUILD_PLUGIN_TGEO=ON -DACTS_BUILD_PLUGIN_DD4HEP=ON -DACTS_BUILD_PLUGIN_DIGITIZATION=ON -DACTS_BUILD_PLUGIN_JSON=ON -DACTS_BUILD_PLUGIN_ACTSVG=ON '
        self.config['cxx_standard'] = 17

    def setup(self, db):
        # ACTS require C++14 (at least). We  check that it is set
        if int(self.config['cxx_standard']) < 17:
            message = "ERROR. cxx_standard must be 17 or above to build ACTS.\n"\
                      "To set cxx_standard globally:\n"\
                      "   edpm config global cxx_standard=17\n"\
                      "To set cxx_standard for acts:\n"\
                      "   edpm config acts cxx_standard=17\n"\
                      "(!) Make sure cmake is regenerated after. (rm <top_dir>/acts and run edpm install acts again)\n"
            raise ValueError(message)

        print(self.config['branch'])
        # Call GitCmakeRecipe `default` setup function
        super(ActsRecipe, self).setup(db)

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        path = data['install_path']

        yield Set('ACTS_DIR', path)

        # it could be lib or lib64. There are bugs on different platforms (RHEL&centos and WSL included)
        # https://stackoverflow.com/questions/46847939/config-site-for-vendor-libs-on-centos-x86-64
        # https: // bugzilla.redhat.com / show_bug.cgi?id = 1510073

        import platform
        if platform.system() == 'Darwin':
            yield Append('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        yield Append('LD_LIBRARY_PATH', os.path.join(path, 'lib'))

        # share/cmake/Acts
        yield Append('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake', 'Acts'))
        yield Append('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake', 'nlohmann_json'))
        yield Append('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake', 'ActsDD4hep'))


    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu18': "libboost-dev libboost-filesystem-dev libboost-program-options-dev libboost-test-dev",
            'ubuntu22': "libboost-dev libboost-filesystem-dev libboost-program-options-dev libboost-test-dev",
            'centos7': "boost-devel",
            'centos8': "boost-devel",
        },
    }


import os

import click

from edpm.engine.api import pass_edpm_context, edpmApi
from edpm.engine.db import INSTALL_PATH, IS_OWNED, SOURCE_PATH, BUILD_PATH
from edpm.engine.output import markup_print as mprint


# @click.group(invoke_without_command=True)
@click.command()
@click.argument('packet_names', nargs=-1, metavar='<packet-name>')
@click.option('--install', 'print_mode', flag_value='install', help="Print installation directory")
@click.option('--build', 'print_mode', flag_value='build', help="Print build directory")
@click.option('--source', 'print_mode', flag_value='source', help="Print source directory")
@click.option('--all', 'print_mode', default=True, flag_value='all', help="Print all information on the package")


@pass_edpm_context
@click.pass_context
def pwd(ctx, ectx, packet_names, print_mode):
    """Shows a directory related to the active packet

    Remark:
        Running `edpm` command shows the directories too, but this command shows the dir
        for a package and can be used in scripts and whatsoever

    Usage:
        edpm cd                  # cd-s to edpm top dir
        edpm cd <packet-name>    # cd-s to <packet-name> installation dir

    """
    from edpm.engine.db import INSTALL_PATH, IS_OWNED, SOURCE_PATH, BUILD_PATH
    assert isinstance(ectx, edpmApi)

    # We need DB ready for this cli command
    ectx.ensure_db_exists()

    # Check that the packet name is from known packets
    if not packet_names:
        pwd_path = ectx.db.top_dir
        mprint(pwd_path)

    else:
        # Get installation data for the active install
        packet_name = packet_names[0]
        ectx.ensure_installer_known(packet_name)
        install_data = ectx.db.get_active_install(packet_name)
        if not install_data:
            mprint("No active installation data found for the packet '{}'", packet_name)
            raise click.Abort()

        if print_mode == 'install':
            _print_single_path(install_data, INSTALL_PATH)
        elif print_mode == 'build':
            _print_single_path(install_data, BUILD_PATH)
        elif print_mode == 'source':
            _print_single_path(install_data, SOURCE_PATH)
        else:
            _pretty_print_all(install_data)


def _print_single_path(install_data, what_path):
    if what_path in install_data:
        print(install_data[what_path])


def _pretty_print_all(install_data):
    # If edpm 'owns' an installation at this point we assume that parent directory usually is needed
    # Like if edpm owns 'root' it is assumed that <top-dir>/root is needed rather than <top-dir>/root/root-v6-...
    pwd_path = install_data[INSTALL_PATH]
    mprint("<blue><b>Install path: </b></blue>\n{}", pwd_path)

    if install_data[IS_OWNED]:
        pwd_path = os.path.dirname(pwd_path)
        mprint("<blue><b>edpm 'owned' base path: </b></blue>\n{}", pwd_path)

    # Source path if known
    if SOURCE_PATH in install_data:
        mprint("<blue><b>Sources: </b></blue>\n{}", install_data[SOURCE_PATH])

    if BUILD_PATH in install_data:
        mprint("<blue><b>Build dir: </b></blue>\n{}", install_data[BUILD_PATH])

    # os.chdir(cd_path)







import inspect
import io
import os
import click
import appdirs


from edpm.engine.db import PacketStateDatabase, merge_db
from edpm.engine.recipe import Recipe
from edpm.engine.recipe_manager import RecipeManager
from edpm.engine.output import markup_print as mprint
from edpm.engine.py23 import to_unicode


edpm_HOME_PATH = 'edpm_home_path'       # Home path of the edpm
edpm_DATA_PATH = 'edpm_data_path'       # Path where
DB_FILE_PATH = 'db_file_path'           # Database file path
ENV_SH_PATH = 'env_sh_path'             # SH environment generated file path
ENV_CSH_PATH = 'env_csh_path'           # CSH environment generated file path


class edpmApi(object):
    """This class holds data that is provided to most edpm CLI commands"""

    def __init__(self):
        self.db = PacketStateDatabase()
        self.pm = RecipeManager()
        self.config = {}
        self._set_paths()

    def load_shmoad_ugly_toad(self):
        """Load the state from disk. DB and packet installers"""

        db_existed = self.load_db_if_exists()   # Try to load the DB

        self.pm.load_installers()               # load installers

        # Update DB "known installers"
        self.db.known_packet_names = self.pm.recipes_by_name.keys()

        return db_existed

    def load_db_if_exists(self):
        if self.db.exists():
            self.db.load()
            return True
        return False

    def merge_external_db(self, file_path):
        import_db = PacketStateDatabase()
        import_db.file_path = file_path
        import_db.load()
        merge_db(self.db, import_db)

    def ensure_db_exists(self):
        """Check if DB exist, create it or aborts everything

         All ensure_xxx functions check the problem, fix it or write message and call Click.Abort()
         """
        # save DB no db...
        if not self.db.exists():
            mprint("<green>creating database...</green>")
            self.db.save()

    def ensure_installer_known(self, packet_name):
        """Check if packet_name is of known packets or aborts everything

           All ensure_xxx functions check the problem, fix it or write message and call Click.Abort()
        """

        # Check if packet_name is all, missing or for known packet
        names = [to_unicode(n) for n in self.pm.recipes_by_name.keys()]
        is_valid_packet_name = to_unicode(packet_name) in names

        if not is_valid_packet_name:
            print("Packet with name '{}' is not found".format(packet_name))  # don't know what to do
            raise click.Abort()

    def save_shell_environ(self, file_path, shell):
        """Generates and saves shell environment to a file
        :param file_path: Path to file
        :param shell: 'bash' or 'csh'
        """

        with io.open(file_path, 'w', encoding='utf8') as outfile:    # Write file
            outfile.write(to_unicode(self.pm.gen_shell_env_text(self.db.get_active_installs(), shell=shell)))

    def save_default_bash_environ(self):
        """Generates and saves bash environment to a default file path"""
        self.save_shell_environ(self.config[ENV_SH_PATH], 'bash')

    def save_default_csh_environ(self):
        """Generates and saves csh/tcsh environment to a default file path"""
        self.save_shell_environ(self.config[ENV_CSH_PATH], 'csh')

    def _set_paths(self):
        """Sets edpm paths"""

        # edpm home path
        # call 'dirname' 3 times as we have <db path>/edpm/cli
        edpm_home_path = os.path.dirname(os.path.dirname(os.path.dirname(inspect.stack()[0][1])))
        self.config[edpm_HOME_PATH] = edpm_home_path

        #
        # edpm data path. It is where db.json and environment files are located
        # We try to read it from edpm_DATA_PATH environment variable and then use standard (XDG) location
        edpm_data_path = os.environ.get('edpm_DATA_PATH', None)
        if not edpm_data_path:
            # Get the default (XDG or whatever) standard path to store user data
            edpm_data_path = appdirs.user_data_dir("edpm", "Dmitry Romanov")

            # In this case we care and create the directory
            if not os.path.isdir(edpm_data_path):
                from edpm.engine.commands import run
                run('mkdir -p "{}"'.format(edpm_data_path))

        self.config[edpm_DATA_PATH] = edpm_data_path

        #
        # Database path
        self.db.file_path = os.path.join(edpm_data_path, "db.json")
        self.config[DB_FILE_PATH] = self.db.file_path

        #
        # environment script paths
        self.config[ENV_SH_PATH] = os.path.join(edpm_data_path, "env.sh")
        self.config[ENV_CSH_PATH] = os.path.join(edpm_data_path, "env.csh")

    def configure_recipes(self):
        config = {}
        config.update(self.db.get_global_config())
        for recipe in self.pm.recipes_by_name.values():
            assert isinstance(recipe, Recipe)
            recipe.config.update(config)

    def update_python_env(self, process_chain=(), mode=''):
        """Update python os.environ assuming we will install missing packets

           This func gets all 'active installation' of packages out of db to build environment
           if process_chain and mode are given, then depending on 'mode':
              'missing' : replace missing installations assuming we will install the package
              'all'     : replace all packets installation path assuming we will install all by our script
              ''        : just skip missing
        """

        if process_chain is None:
            process_chain = []
        from edpm.engine.db import IS_OWNED, IS_ACTIVE, INSTALL_PATH

        # Pretty header
        mprint("\n")
        mprint("<magenta>=========================================</magenta>")
        mprint("<green> SETTING ENVIRONMENT</green>")
        mprint("<magenta>=========================================</magenta>\n")

        # 1st, we need all data we have. All because installing packets can utilize the fa
        packet_data_by_name = {name: inst for name, inst in self.db.get_active_installs().items() if inst}

        # if we have process_chain
        for request in process_chain:
            packet_data = None
            if mode == 'missing' and (request.name not in packet_data_by_name.keys()):
                # There is no installation data for the packet, but we assume we will install it now!
                packet_data = {
                    INSTALL_PATH: request.recipe.install_path,
                    IS_ACTIVE: True,
                    IS_OWNED: True
                }
            elif mode == 'all':
                # We overwrite installation path for the packet
                packet_data = {
                    INSTALL_PATH: request.recipe.install_path,
                    IS_ACTIVE: True,
                    IS_OWNED: True
                }
            if packet_data:
                packet_data_by_name[request.name] = packet_data

        for name, data in packet_data_by_name.items():
            if name not in self.pm.env_generators.keys():  # Skip if we don't know environment generator for a packet
                continue
            # If we have a generator for this program and installation data
            mprint("<blue><b>Updating python environment for '{}'</b></blue>".format(name))
            generators = self.pm.env_generators[name]
            for step in generators(packet_data_by_name[name]):  # Go through 'env generators' look engine/env_gen.py
                step.update_python_env()  # Do environment update

    def req_get_known_os(self):
        names_set = set()
        for name in self.pm.recipes_by_name.keys():
            if 'required' in self.pm.os_deps_by_name[name]:
                for key in self.pm.os_deps_by_name[name]['required'].keys():
                    names_set.add(key)
        return names_set   # TODO change for

    def req_get_deps(self, os_name, packet_names=None):
        """ Returns
        # No packets provided, so all packets are selected

        :param os_name: Name of os like centos, ubuntu
        :param packet_names: If set, returns requirements for this packages and their deps. If None - all packages deps
        :return: list of required and optional packages
        """

        # Step 0 - What exactly packets are involved?
        if packet_names:
            names_with_deps = []    # Names of packets and their dependencies

            # get all dependencies
            for packet_name in packet_names:
                self.ensure_installer_known(packet_name)
                names_with_deps += self.pm.get_installation_chain_names(packet_name)  # func returns name + its_deps

            names_with_deps = list(set(names_with_deps))  # remove repeating names

        else:
            # No packets provided, so all packets are selected
            names_with_deps = self.pm.recipes_by_name.keys()

        # Step 1 - Form results
        required = []
        optional = []
        for name in names_with_deps:
            if os_name in self.pm.os_deps_by_name[name]['required'].keys():
                required.extend(self.pm.os_deps_by_name[name]['required'][os_name].replace(',', ' ').split())
            if os_name in self.pm.os_deps_by_name[name]['optional'].keys():
                optional.extend(self.pm.os_deps_by_name[name]['optional'][os_name].replace(',', ' ').split())

        # remove emtpy elements and repeating elements
        required = list(set([r for r in required if r]))
        optional = list(set([o for o in optional if o]))

        return required, optional


# Create a database class and @pass_db decorator so our commands could use it
pass_edpm_context = click.make_pass_decorator(edpmApi, ensure=True)


def print_packets_info(db):
    """Prints known installations of packets and what packet is selected"""

    from edpm.engine.db import IS_OWNED, IS_ACTIVE, INSTALL_PATH
    assert (isinstance(db, PacketStateDatabase))

    installed_names = [name for name in db.packet_names]

    # Fancy print of installed packets
    if installed_names:
        mprint('\n<b><magenta>INSTALLED PACKETS:</magenta></b> (*-active):')
        for packet_name in installed_names:
            mprint(' <b><blue>{}</blue></b>:'.format(packet_name))
            installs = db.get_installs(packet_name)
            for i, installation in enumerate(installs):
                is_owned_str = '<green>(owned)</green>' if installation[IS_OWNED] else ''
                is_active = installation[IS_ACTIVE]
                is_active_str = '*' if is_active else ' '
                path_str = installation[INSTALL_PATH]
                id_str = "[{}]".format(i).rjust(4) if len(installs) > 1 else ""
                mprint("  {}{} {} {}".format(is_active_str, id_str, path_str, is_owned_str))

    not_installed_names = [name for name in db.known_packet_names if name not in installed_names]

    # Fancy print of installed packets
    if not_installed_names:
        mprint("\n<b><magenta>NOT INSTALLED:</magenta></b>\n(could be installed by 'edpm install')")
        for packet_name in not_installed_names:
            mprint(' <b><blue>{}</blue></b>'.format(packet_name))